from django.db import models
from django.utils import translation

from .fields import WikidataEntityField
from . import client


def get_string(value):
    try:
        lang = translation.get_language()
        return value[lang]
    except KeyError:
        pass
    return value


class DefaultWikidataMeta:
    field_mapping = None
    constraints = None


class WikidataProxy:
    def __init__(self, instance):
        self.instance = instance

    @property
    def options(self):
        if hasattr(self.instance, 'WikidataMeta'):
            return self.instance.WikidataMeta
        return DefaultWikidataMeta

    @property
    def model(self):
        return type(self.instance)

    def collect_fields(self):
        self.fields = {}
        for field in self.model._meta.get_fields():
            if isinstance(field, WikidataEntityField):
                self.fields[field.name] = field

    def update_fields(self):
        if self.options.field_mapping is None:
            return
        for field_name, definition in self.options.field_mapping:
            self.update_field(field_name, definition)

    def update_field(self, field_name, definition):
        value = self.get_mapping_value(definition)
        update_field = self.model._meta.get_field(field_name)
        update_value = self.get_update_field_value(update_field, value)
        setattr(self.instance, field_name, update_value)

    def get_update_field_value(self, update_field, value):
        if isinstance(update_field, (models.TextField, models.CharField)):
            return get_string(value)
        return value

    def get_mapping_value(self, definition):
        entity_field, accessor = definition
        entity = getattr(self.instance, entity_field)
        if isinstance(accessor, str):
            prop = client.get(accessor)
            return entity[prop]
        return accessor(entity)


class WikidataMixin:
    @property
    def wikidata(self):
        if not hasattr(self, '_wikidataproxy'):
            self._wikidataproxy = WikidataProxy(self)
        return self._wikidataproxy



from django.db import models
from django.core import exceptions

from wikidata.entity import Entity

from . import client
from .forms import WikidataItemFormField
from .widgets import WikidataItemWidget
from .validators import validate_wikidata_item


def make_entity(value):
    validate_wikidata_item(value)
    return client.get(value)


class EntityDescriptor:
    def __init__(self, field):
        self.field = field

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        if self.field.name in instance.__dict__:
            entity = instance.__dict__[self.field.name]
        else:
            instance.refresh_from_db(fields=[self.field.name])
            entity = getattr(instance, self.field.name)

        if not entity:
            instance.__dict__[self.field.name] = None

        elif isinstance(entity, str):
            attr = make_entity(entity)
            instance.__dict__[self.field.name] = attr

        # That was fun, wasn't it?
        return instance.__dict__[self.field.name]

    def __set__(self, instance, value):
        if not value:
            instance.__dict__[self.field.name] = None
        elif isinstance(value, str):
            attr = make_entity(value)
            instance.__dict__[self.field.name] = attr
        elif isinstance(value, Entity):
            instance.__dict__[self.field.name] = value
        else:
            raise ValueError('Cannot assign "{}" to Wikidata Item Field'.format(value))


class WikidataEntityField(models.CharField):
    descriptor_class = EntityDescriptor
    empty_values = ()

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 50)
        super().__init__(*args, **kwargs)
        self.validators = []  # FIXME: validate length of QID

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)
        setattr(cls, self.name, self.descriptor_class(self))

    def validate(self, value, model_instance):
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if value is None and not self.null:
            raise exceptions.ValidationError(self.error_messages['null'], code='null')

        if not self.blank and value is None:
            raise exceptions.ValidationError(self.error_messages['blank'], code='blank')

    def from_db_value(self, value, expression, connection):
        if not value:
            return None
        return client.get(value)

    def to_python(self, value):
        if isinstance(value, Entity):
            return value

        if value is None:
            return value

        return client.get(value)

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        value = self.to_python(value)
        return value.id

    def formfield(self, **kwargs):
        # This is a fairly standard way to set up some defaults
        # while letting the caller override them.
        defaults = dict(kwargs)
        defaults.update({
            'form_class': WikidataItemFormField,
            'widget': WikidataItemWidget,
        })
        return super().formfield(**defaults)

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        return self.get_prep_value(value)


class WikidataItemField(WikidataEntityField):
    pass



from django import forms


def validate_wikidata_item(value):
    if not value.startswith('Q'):
        raise forms.ValidationError('"{}" is not a Wikidata Item ID'.format(value))
    try:
        int(value[1:])
    except ValueError:
        raise forms.ValidationError('"{}" is not a Wikidata Item ID'.format(value))



from django.forms import widgets


class WikidataItemWidget(widgets.TextInput):
    def __init__(self, attrs=None):
        if attrs is not None:
            attrs = attrs.copy()
        if attrs is None:
            attrs = {}
        attrs.setdefault('pattern', 'Q\d+')
        super().__init__(attrs)

    def format_value(self, value):
        """
        Return a value as it should appear when rendered in a template.
        """
        if not value:
            return None
        return value



from django import forms

from .widgets import WikidataItemWidget
from .validators import validate_wikidata_item
from . import client


class WikidataItemFormField(forms.CharField):
    widget = WikidataItemWidget

    def to_python(self, value):
        """Normalize data to a list of strings."""
        # Return an empty list if no input was given.
        if not value:
            return None
        if isinstance(value, str):
            validate_wikidata_item(value)
            return client.get(value)
        return value

    def clean(self, value):
        """Check if value consists only of valid emails."""
        value = self.to_python(value)
        if value is None and self.required:
            raise forms.ValidationError(self.error_messages['required'], code='required')
        return value

    def prepare_value(self, value):
        return value.id
